@@ -44,6 +44,16 @@ static DEFINE_SEMAPHORE(core_lock);
 struct bus_type ddcci_bus_type;
 EXPORT_SYMBOL_GPL(ddcci_bus_type);
 
+/* Assert neccessary string array sizes  */
+#ifndef sizeof_field
+# define sizeof_field(t,m) FIELD_SIZEOF(t,m)
+#endif
+static_assert(sizeof_field(struct ddcci_device, prot) > 8);
+static_assert(sizeof_field(struct ddcci_device, type) > 8);
+static_assert(sizeof_field(struct ddcci_device, model) > 8);
+static_assert(sizeof_field(struct ddcci_device, vendor) > 8);
+static_assert(sizeof_field(struct ddcci_device, module) > 8);
+
 /* Internal per-i2c-client driver data */
 struct ddcci_bus_drv_data {
 	unsigned long quirks;
@@ -52,9 +62,23 @@ struct ddcci_bus_drv_data {
 	unsigned char recv_buffer[DDCCI_RECV_BUFFER_SIZE];
 };
 
+/* Replace non-alphanumeric characters in a string (used for modalias) */
+static void ddcci_modalias_clean(char *string, size_t n, char replacement)
+{
+	int i;
+	for (i = 0; i < n; ++i) {
+		char c = string[i];
+		if (c == 0) {
+			return;
+		} else if (c < '0' || (c > '9' && c < 'A') || (c > 'Z' && c < 'a') || c > 'z') {
+			string[i] = replacement;
+		}
+	}
+}
+
 /* Write a message to the DDC/CI bus using i2c_smbus_write_byte() */
 static int __ddcci_write_bytewise(struct i2c_client *client, unsigned char addr,
-				  bool p_flag, const unsigned char *buf,
+				  bool p_flag, const unsigned char * __restrict buf,
 				  unsigned char len)
 {
 	int ret = 0;
@@ -301,7 +325,7 @@ static int ddcci_get_caps(struct i2c_client *client, unsigned char addr,
 			  unsigned char *buf, unsigned int len)
 {
 	int result = 0, counter = 0, offset = 0;
-	unsigned char cmd[3] = { DDCCI_COMMAND_CAPS, 0x00, 0x00};
+	unsigned char cmd[3] = { DDCCI_COMMAND_CAPS, 0x00, 0x00 };
 	unsigned char *chunkbuf = kzalloc(35, GFP_KERNEL);
 
 	if (!chunkbuf)
@@ -329,7 +353,11 @@ static int ddcci_get_caps(struct i2c_client *client, unsigned char addr,
 				result = -EIO;
 				goto err_free;
 			}
-			memcpy(buf, chunkbuf+3, result-3);
+			if (result < 3) {
+				result = -EIO;
+				goto err_free;
+			}
+			memcpy(buf, chunkbuf+3, min((unsigned int)result-3, len));
 
 			counter++;
 			/* adjust offset, etc. */
@@ -364,7 +392,7 @@ static int ddcci_identify_device(struct i2c_client *client, unsigned char addr,
 {
 	int i, payload_len, ret = -ENODEV;
 	unsigned long quirks;
-	unsigned char cmd[2] = { DDCCI_COMMAND_ID, 0x00 };
+	unsigned char cmd[1] = { DDCCI_COMMAND_ID };
 	unsigned char *buffer;
 	unsigned char xor = DDCCI_HOST_ADDR_EVEN;
 	struct ddcci_bus_drv_data *bus_drv_data;
@@ -375,7 +403,7 @@ static int ddcci_identify_device(struct i2c_client *client, unsigned char addr,
 
 	/* Send Identification command */
 	if (!(quirks & DDCCI_QUIRK_WRITE_BYTEWISE)) {
-		ret = __ddcci_write_block(client, addr, buffer, true, cmd, 1);
+		ret = __ddcci_write_block(client, addr, buffer, true, cmd, sizeof(cmd));
 		dev_dbg(&client->dev,
 			"[%02x:%02x] writing identification command in block mode: %d\n",
 			client->addr << 1, addr, ret);
@@ -386,12 +414,12 @@ static int ddcci_identify_device(struct i2c_client *client, unsigned char addr,
 			dev_info(&client->dev,
 				"DDC/CI bus quirk detected: writes must be done bytewise\n");
 			/* Some devices need writing twice after a failed blockwise write */
-			__ddcci_write_bytewise(client, addr, true, cmd, 2);
+			__ddcci_write_bytewise(client, addr, true, cmd, sizeof(cmd));
 			msleep(delay);
 		}
 	}
 	if (ret < 0 && (quirks & DDCCI_QUIRK_WRITE_BYTEWISE)) {
-		ret = __ddcci_write_bytewise(client, addr, true, cmd, 2);
+		ret = __ddcci_write_bytewise(client, addr, true, cmd, sizeof(cmd));
 		dev_dbg(&client->dev,
 			"[%02x:%02x] writing identification command in bytewise mode: %d\n",
 			client->addr << 1, addr, ret);
@@ -736,7 +764,7 @@ static ssize_t ddcci_attr_prot_show(struct device *dev,
 	size_t len;
 
 	if (likely(device != NULL)) {
-		len = strlen(device->prot);
+		len = strnlen(device->prot, sizeof(device->prot));
 		strncpy(buf, device->prot, PAGE_SIZE);
 		if (len == 0) {
 			ret = len;
@@ -758,7 +786,7 @@ static ssize_t ddcci_attr_type_show(struct device *dev,
 	size_t len;
 
 	if (likely(device != NULL)) {
-		len = strlen(device->type);
+		len = strnlen(device->type, sizeof(device->type));
 		strncpy(buf, device->type, PAGE_SIZE);
 		if (len == 0) {
 			ret = len;
@@ -780,7 +808,7 @@ static ssize_t ddcci_attr_model_show(struct device *dev,
 	size_t len;
 
 	if (likely(device != NULL)) {
-		len = strlen(device->model);
+		len = strnlen(device->model, sizeof(device->model));
 		strncpy(buf, device->model, PAGE_SIZE);
 		if (len == 0) {
 			ret = len;
@@ -802,7 +830,7 @@ static ssize_t ddcci_attr_vendor_show(struct device *dev,
 	size_t len;
 
 	if (likely(device != NULL)) {
-		len = strlen(device->vendor);
+		len = strnlen(device->vendor, sizeof(device->vendor));
 		strncpy(buf, device->vendor, PAGE_SIZE);
 		if (len == 0) {
 			ret = len;
@@ -824,7 +852,7 @@ static ssize_t ddcci_attr_module_show(struct device *dev,
 	size_t len;
 
 	if (likely(device != NULL)) {
-		len = strlen(device->module);
+		len = strnlen(device->module, sizeof(device->module));
 		strncpy(buf, device->module, PAGE_SIZE);
 		if (len == 0) {
 			ret = len;
@@ -850,6 +878,35 @@ static ssize_t ddcci_attr_serial_show(struct device *dev,
 	return ret;
 }
 
+static ssize_t ddcci_attr_modalias_show(struct device *dev,
+				      struct device_attribute *attr, char *buf)
+{
+	struct ddcci_device *device = ddcci_verify_device(dev);
+	ssize_t ret = -ENOENT;
+	char model[ARRAY_SIZE(device->model)];
+	char vendor[ARRAY_SIZE(device->model)];
+	char module[ARRAY_SIZE(device->model)];
+
+	if (likely(device != NULL)) {
+		memcpy(model, device->model, sizeof(model));
+		memcpy(vendor, device->vendor, sizeof(vendor));
+		memcpy(module, device->module, sizeof(module));
+		ddcci_modalias_clean(model, sizeof(model), '_');
+		ddcci_modalias_clean(vendor, sizeof(vendor), '_');
+		ddcci_modalias_clean(module, sizeof(module), '_');
+
+		ret = scnprintf(buf, PAGE_SIZE, "%s%s-%s-%s-%s-%s\n",
+			DDCCI_MODULE_PREFIX,
+			device->prot,
+			device->type,
+			model,
+			vendor,
+			module
+		);
+	}
+	return ret;
+}
+
 static DEVICE_ATTR(capabilities, S_IRUGO, ddcci_attr_capabilities_show, NULL);
 static DEVICE_ATTR(idProt, S_IRUGO, ddcci_attr_prot_show, NULL);
 static DEVICE_ATTR(idType, S_IRUGO, ddcci_attr_type_show, NULL);
@@ -857,6 +914,7 @@ static DEVICE_ATTR(idModel, S_IRUGO, ddcci_attr_model_show, NULL);
 static DEVICE_ATTR(idVendor, S_IRUGO, ddcci_attr_vendor_show, NULL);
 static DEVICE_ATTR(idModule, S_IRUGO, ddcci_attr_module_show, NULL);
 static DEVICE_ATTR(idSerial, S_IRUGO, ddcci_attr_serial_show, NULL);
+static DEVICE_ATTR(modalias, S_IRUGO, ddcci_attr_modalias_show, NULL);
 
 static struct attribute *ddcci_char_device_attrs[] = {
 	&dev_attr_capabilities.attr,
@@ -866,6 +924,7 @@ static struct attribute *ddcci_char_device_attrs[] = {
 	&dev_attr_idVendor.attr,
 	&dev_attr_idModule.attr,
 	&dev_attr_idSerial.attr,
+	&dev_attr_modalias.attr,
 	NULL,
 };
 ATTRIBUTE_GROUPS(ddcci_char_device);
@@ -875,14 +934,24 @@ ATTRIBUTE_GROUPS(ddcci_char_device);
 static int ddcci_device_uevent(struct device *dev, struct kobj_uevent_env *env)
 {
 	struct ddcci_device	*device = to_ddcci_device(dev);
+	char model[ARRAY_SIZE(device->model)];
+	char vendor[ARRAY_SIZE(device->vendor)];
+	char module[ARRAY_SIZE(device->module)];
+
+	memcpy(model, device->model, sizeof(model));
+	memcpy(vendor, device->vendor, sizeof(vendor));
+	memcpy(module, device->module, sizeof(module));
+	ddcci_modalias_clean(model, sizeof(model), '_');
+	ddcci_modalias_clean(vendor, sizeof(vendor), '_');
+	ddcci_modalias_clean(module, sizeof(module), '_');
 
 	if (add_uevent_var(env, "MODALIAS=%s%s-%s-%s-%s-%s",
 			   DDCCI_MODULE_PREFIX,
 			   device->prot,
 			   device->type,
-			   device->model,
-			   device->vendor,
-			   device->module
+			   model,
+			   vendor,
+			   module
 		))
 		return -ENOMEM;
 
@@ -1259,15 +1328,16 @@ static char *ddcci_capstr_tok(const char *s, int depth)
  * If the tag is not found or another error occurs, an ERR_PTR is returned
  * and `length` stays untouched.
  */
-const char *ddcci_find_capstr_item(const char *capabilities, const char *tag,
-				    ptrdiff_t *length)
+const char *ddcci_find_capstr_item(const char * capabilities,
+				   const char * __restrict tag,
+				   size_t *length)
 {
 	const char *src = capabilities, *ptr;
 	ptrdiff_t len;
-	int taglen = strlen(tag);
+	int taglen = strnlen(tag, 1024);
 
 	/* Check length of requested tag */
-	if (unlikely(taglen <= 0 || taglen > 65535))
+	if (unlikely(taglen <= 0 || taglen > 1023))
 		return ERR_PTR(-EINVAL);
 
 	/* Find tag */
@@ -1289,17 +1359,17 @@ const char *ddcci_find_capstr_item(const char *capabilities, const char *tag,
 
 	/* Return pointer and length */
 	if (likely(length != NULL))
-		*length = len;
+		*length = (size_t)len;
 	return src;
 }
 EXPORT_SYMBOL(ddcci_find_capstr_item);
 
 /* Search the capability string for a tag and copy the value to dest */
 static int ddcci_cpy_capstr_item(char *dest, const char *src,
-				  const char *tag, size_t maxlen)
+				  const char * __restrict tag, size_t maxlen)
 {
 	const char *ptr;
-	ptrdiff_t len;
+	size_t len;
 
 	/* Find tag */
 	ptr = ddcci_find_capstr_item(src, tag, &len);
@@ -1308,7 +1378,7 @@ static int ddcci_cpy_capstr_item(char *dest, const char *src,
 	}
 
 	/* Copy value */
-	memcpy(dest, ptr, (len < maxlen) ? len : maxlen);
+	memcpy(dest, ptr, min(len, maxlen));
 	return 0;
 }
 
@@ -1326,32 +1396,51 @@ static int ddcci_parse_capstring(struct ddcci_device *device)
 		return -EINVAL;
 
 	/* get prot(...) */
-	ret = ddcci_cpy_capstr_item(device->prot, capstr, "prot", 8);
-	if (ret)
-		return ret;
+	ret = ddcci_cpy_capstr_item(device->prot, capstr, "prot", sizeof(device->prot)-1);
+	if (ret) {
+		if (ret == -ENOENT) {
+			dev_warn(&device->dev, "malformed capability string: no protocol tag");
+			memset(device->prot, 0, sizeof(device->prot)-1);
+		} else {
+			return ret;
+		}
+	}
 
 	/* get type(...) */
-	ret = ddcci_cpy_capstr_item(device->type, capstr, "type", 8);
+	ret = ddcci_cpy_capstr_item(device->type, capstr, "type", sizeof(device->type)-1);
 	if (ret) {
 		if (ret == -ENOENT) {
 			dev_warn(&device->dev, "malformed capability string: no type tag");
-			memset(device->type, 0, 8);
+			memset(device->type, 0, sizeof(device->type)-1);
 		} else {
 			return ret;
 		}
 	}
 
 	/* and then model(...) */
-	ret = ddcci_cpy_capstr_item(device->model, capstr, "model", 8);
+	ret = ddcci_cpy_capstr_item(device->model, capstr, "model", sizeof(device->model)-1);
 	if (ret) {
 		if (ret == -ENOENT) {
 			dev_warn(&device->dev, "malformed capability string: no model tag");
-			memset(device->model, 0, 8);
+			memset(device->model, 0, sizeof(device->model)-1);
 		} else {
 			return ret;
 		}
 	}
 
+	/* if there is no protocol tag */
+	if (!device->prot[0]) {
+		/* and no type tag: give up. */
+		if (!device->type[0])
+			return -ENOENT;
+
+		/* Assume protocol "monitor" if type is "LCD" or "CRT" */
+		if (strncasecmp(device->type, "LCD", sizeof(device->type)-1) == 0
+		 || strncasecmp(device->type, "CRT", sizeof(device->type)-1) == 0) {
+			memcpy(device->prot, "monitor", 7);
+		}
+	}
+
 	/* skip the rest for now */
 
 	return 0;
@@ -1373,14 +1462,14 @@ static int ddcci_detect_device(struct i2c_client *client, unsigned char addr,
 	buffer = kmalloc(16384, GFP_KERNEL);
 	if (!buffer) {
 		ret = -ENOMEM;
-		goto end;
+		goto err_end;
 	}
 
 	/* Allocate device struct */
 	device = kzalloc(sizeof(struct ddcci_device), GFP_KERNEL);
 	if (!device) {
 		ret = -ENOMEM;
-		goto end;
+		goto err_end;
 	}
 
 	/* Initialize device */
@@ -1418,8 +1507,14 @@ static int ddcci_detect_device(struct i2c_client *client, unsigned char addr,
 
 	/* Read identification and check for quirks */
 	ret = ddcci_identify_device(client, addr, buffer, 29);
-	if (ret < 0)
-		goto err_free;
+	if (ret < 0) {
+		if (!dependent && (ret == -EBADMSG || ret == -EMSGSIZE)) {
+			dev_warn(&device->dev, "DDC/CI main device sent broken response on identification. Trying to detect solely based on capability information.\n");
+		} else {
+			goto err_free;
+		}
+	}
+
 	if (ret == 29 && buffer[0] == DDCCI_REPLY_ID) {
 		memcpy(device->vendor, &buffer[7], 8);
 		memcpy(device->module, &buffer[17], 8);
@@ -1429,17 +1524,37 @@ static int ddcci_detect_device(struct i2c_client *client, unsigned char addr,
 	/* Read capabilities */
 	ret = ddcci_get_caps(client, addr, buffer, 16384);
 	if (ret > 0) {
-		device->capabilities = kzalloc(ret+1, GFP_KERNEL);
-		if (!device->capabilities) {
-			ret = -ENOMEM;
-			goto err_free;
+		/* Fixup unparenthesized capability strings, but only if the first
+		   character is an ascii lower case letter.
+		   This should still allow an early exit for completely garbled
+		   data but helps detecting devices where only the parentheses are
+		   missing, as the second char must be the first character of a
+		   keyword. */
+		if (ret > 2 && buffer[0] >= 'a' && buffer[0] <= 'z') {
+			dev_err(&device->dev, "DDC/CI device quirk detected: unparenthesized capability string\n");
+			device->capabilities = kzalloc(ret+3, GFP_KERNEL);
+			if (!device->capabilities) {
+				ret = -ENOMEM;
+				goto err_free;
+			}
+			device->capabilities_len = ret+2;
+			memcpy(&(device->capabilities[1]), buffer, ret);
+			device->capabilities[0] = '(';
+			device->capabilities[ret+1] = ')';
+		} else {
+			/* Standard case: simply copy the received string */
+			device->capabilities = kzalloc(ret+1, GFP_KERNEL);
+			if (!device->capabilities) {
+				ret = -ENOMEM;
+				goto err_free;
+			}
+			device->capabilities_len = ret;
+			memcpy(device->capabilities, buffer, ret);
 		}
-		device->capabilities_len = ret;
-		memcpy(device->capabilities, buffer, ret);
 
 		ret = ddcci_parse_capstring(device);
 		if (ret) {
-			dev_err(&device->dev, "malformed capability string: %d", ret);
+			dev_err(&device->dev, "malformed capability string: \"%s\" errno %d\n", device->capabilities, ret);
 			ret = -EINVAL;
 			goto err_free;
 		}
@@ -1457,10 +1572,12 @@ static int ddcci_detect_device(struct i2c_client *client, unsigned char addr,
 	/* Setup chardev */
 	down(&core_lock);
 	ret = ddcci_setup_char_device(device);
+	up(&core_lock);
 	if (ret)
 		goto err_free;
 
-	/* Add device */
+	/* Release semaphore and add device to the tree */
+	up(&drv_data->sem);
 	pr_debug("found device at %d:%02x:%02x\n", client->adapter->nr, outer_addr, addr);
 	ret = device_add(&device->dev);
 	if (ret)
@@ -1469,10 +1586,10 @@ static int ddcci_detect_device(struct i2c_client *client, unsigned char addr,
 	goto end;
 err_free:
 	put_device(&device->dev);
+err_end:
+	up(&drv_data->sem);
 end:
 	kfree(buffer);
-	up(&drv_data->sem);
-	up(&core_lock);
 	return ret;
 }
 
@@ -1483,7 +1600,10 @@ static int ddcci_detect(struct i2c_client *client, struct i2c_board_info *info)
 	unsigned char outer_addr;
 	unsigned char inner_addr;
 	unsigned char buf[32];
-	unsigned char cmd[2] = { DDCCI_COMMAND_ID, 0x00 };
+	unsigned char cmd_id[1] = { DDCCI_COMMAND_ID };
+	unsigned char cmd_caps[3] = { DDCCI_COMMAND_CAPS, 0x00, 0x00};
+	unsigned char *cmd;
+	unsigned int cmd_len;
 
 	/* Check for i2c_master_* functionality */
 	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
@@ -1491,21 +1611,25 @@ static int ddcci_detect(struct i2c_client *client, struct i2c_board_info *info)
 		return -ENODEV;
 	}
 
-	/* send Identification Request command */
+	/* send Capabilities Request (for main) or Identification Request command (for dependent devices) */
 	outer_addr = client->addr << 1;
 	inner_addr = (outer_addr == DDCCI_DEFAULT_DEVICE_ADDR) ? DDCCI_HOST_ADDR_ODD : outer_addr | 1;
+	cmd = (outer_addr == DDCCI_DEFAULT_DEVICE_ADDR) ? cmd_caps : cmd_id;
+	cmd_len = (outer_addr == DDCCI_DEFAULT_DEVICE_ADDR) ? sizeof(cmd_caps) : sizeof(cmd_id);
 	pr_debug("detecting %d:%02x\n", client->adapter->nr, outer_addr);
 
-	ret = __ddcci_write_block(client, inner_addr, buf, true, cmd, 2);
+	ret = __ddcci_write_block(client, inner_addr, buf, true, cmd, cmd_len);
 
-	if (ret == -ENXIO) {
-		if (!i2c_check_functionality(client->adapter, I2C_FUNC_SMBUS_WRITE_BYTE))
+	if (ret == -ENXIO || ret == -EIO) {
+		if (!i2c_check_functionality(client->adapter, I2C_FUNC_SMBUS_WRITE_BYTE)) {
+			pr_debug("i2c write failed with ENXIO or EIO but bytewise writing is not supported\n");
 			return -ENODEV;
-		pr_debug("i2c write failed with ENXIO, trying bytewise writing\n");
-		ret = __ddcci_write_bytewise(client, inner_addr, true, cmd, 2);
+		}
+		pr_debug("i2c write failed with ENXIO or EIO, trying bytewise writing\n");
+		ret = __ddcci_write_bytewise(client, inner_addr, true, cmd, cmd_len);
 		if (ret == 0) {
 			msleep(delay);
-			ret = __ddcci_write_bytewise(client, inner_addr, true, cmd, 2);
+			ret = __ddcci_write_bytewise(client, inner_addr, true, cmd, cmd_len);
 		}
 	}
 
@@ -1523,7 +1647,7 @@ static int ddcci_detect(struct i2c_client *client, struct i2c_board_info *info)
 
 	/* check response starts with outer addr */
 	if (buf[0] != outer_addr) {
-		pr_debug("detection failed: invalid identification response (%02x != %02x)\n", buf[0], outer_addr);
+		pr_debug("detection failed: invalid %s response (%02x != %02x)\n", (cmd == cmd_id) ? "identification" : "capabilities", buf[0], outer_addr);
 		pr_debug("received message was %*ph \n", ret, buf);
 		return -ENODEV;
 	}
@@ -1756,5 +1880,5 @@ MODULE_PARM_DESC(autoprobe_addrs, "internal dependent device addresses to autopr
 /* Module description */
 MODULE_AUTHOR("Christoph Grenz");
 MODULE_DESCRIPTION("DDC/CI bus driver");
-MODULE_VERSION("0.3.3");
+MODULE_VERSION("0.4.1");
 MODULE_LICENSE("GPL");
